==> basefunctions-0.3.5.1-py3-none-any.whl: basefunctions/threadpool.py <==
"""
# =============================================================================
#
#  Licensed Materials, Property of Ralph Vogl, Munich
#
#  Project : eod2pd
#
#  Copyright (c) by Ralph Vogl
#
#  All rights reserved.
#
#  Description:
#
#  a simple thread pool library to execute commands in a multithreaded
#  environment
#
#  The thread pool lib supports a timeout mechanism when executing the
#  commands. Specify the timeout value in seconds when creating the thread
#  pool. If the execution time of the command exceeds the timeout value,
#  a timout exception will occur and the command will be retried.
#  The number of retries can also be specified when creating the thread pool.
#  If the command wasn't successful after the specified number of retries,
#  the command will be considered as failed. In this case, a message
#  '##FAILURE##' will be put into the output queue and the item will be put
#  back into the input queue. If the command was successful, the result of
#  the command will be put into the output queue.
# =============================================================================
"""

# pylint: disable=W0718

# -------------------------------------------------------------
# IMPORTS
# -------------------------------------------------------------
import atexit
import ctypes
import queue
import threading
import types
from dataclasses import dataclass
from typing import Any, Callable, List

import decouple

# -------------------------------------------------------------
# DEFINITIONS REGISTRY
# -------------------------------------------------------------

# -------------------------------------------------------------
# DEFINITIONS
# -------------------------------------------------------------


# -------------------------------------------------------------
# VARIABLE DEFINTIONS
# -------------------------------------------------------------


# -------------------------------------------------------------
#  CLASS DEFINITIONS
# -------------------------------------------------------------


@dataclass
class ThreadPoolMessage:
    """
    This class represents a message that can be sent to the thread pool.
    """

    type: str
    retry: int = 3
    timeout: int = 5
    content: Any = None
    hook: Callable[[Any, queue.Queue, queue.Queue, Any], None] = None


def create_threadpool_message(
    _type: str, content: Any, retry: int = 3, timeout: int = 5, hook: Callable = None
) -> ThreadPoolMessage:
    """
    Creates a ThreadPoolMessage object with the specified type and content.

    Args:
        _type (str): The type of the message.
        content (Any): The content of the message.
        retry (int): The number of retries for the message.
        timeout (int): The timeout value for the message.
        hook(Callable): The hook function to call after the message was processed.

    Returns:
        ThreadPoolMessage: The created ThreadPoolMessage object.
    """
    return ThreadPoolMessage(_type, retry, timeout, content, hook)


# pylint too-few-public-methods
class ThreadPoolUserObject:
    """
    This class represents a user object which contains the callable_function
    that can be executed by a thread pool.

    Methods
    -------
    callableFunction(outputQueue, item)
        This function will be called from the thread pool in order to execute
        a specific command. The command is provided in the item parameter
        and is application specific. After processing the command, the result
        of the command can be put into the outputQueue.

    Attributes
    ----------
    None
    """

    def callable_function(
        self,
        thread_local_data: Any,
        input_queue: queue.Queue,
        output_queue: queue.Queue,
        message: ThreadPoolMessage,
    ) -> int:
        """
        This function will be called from the thread pool in order to execute
        a specific command. The command is provided in the message parameter
        and is application specific. After processing the command, the result
        of the command can be put into the output_queue. If there are any
        other commands that needs to be executed, they can be put into the
        input_queue.

        Parameters
        ----------
        thread_local_data : Any
            The thread local data.
        input_queue : queue.LifoQueue
            The input queue for receiving tasks.
        output_queue : queue.Queue
            The queue to put the result of the task.
        message : object
            The message containing the command to execute.

        Returns
        -------
        int
            The return code of the task.
            Return 0 if successful, otherwise any other value.
        """
        raise NotImplementedError


class ThreadPool:
    """
    A thread pool implementation for executing tasks concurrently.

    This class manages a pool of threads that can execute tasks in parallel.
    It provides a convenient way to distribute work across multiple threads
    and process the results asynchronously.

    Attributes:
        thread_pool (list): A list of threads in the thread pool.
        input_queue (Queue): The input queue for receiving tasks.
        output_queue (Queue): The output queue for storing results.
        thread_pool_user_objects (dict): A dictionary of thread pool user

    Methods:
        add_thread(target)
            Adds a thread to the thread pool.
        stop_threads()
            Stops all the threads in the thread pool by putting a stop signal
            in the input queue.
        register_message_handler(msg_type, msg_handler)
            Registers a message handler for a specific message type.
        get_message_handler(msg_type)
            Get message handler for a specific message type.
        callback(thread_num, thread_local_data, input_queue, output_queue,
                thread_pool_user_objects)
            Executes the callback function for each item in the input queue.
        get_input_queue()
            Returns the input queue of the thread pool.
        get_output_queue()
            Returns the output queue of the thread pool.
        get_dataframes_from_output_queue()
            Receive a list of DataFrames from the output queue.
    """

    # -------------------------------------------------------------
    # VARIABLE DEFINTIONS
    # -------------------------------------------------------------
    thread_list: List[threading.Thread] = []
    input_queue: queue.Queue = queue.Queue()
    output_queue: queue.Queue = queue.Queue()
    thread_pool_user_objects: dict = {}
    thread_local_data = threading.local()

    # -------------------------------------------------------------
    # INIT METHOD DEFINITION
    # -------------------------------------------------------------
    def __init__(
        self,
        default_thread_pool_user_object: ThreadPoolUserObject,
        num_of_threads: int = 5,
    ) -> None:
        # save the default thread pool user object
        self.thread_pool_user_objects["default"] = (
            default_thread_pool_user_object
            if default_thread_pool_user_object is not None
            else ThreadPoolUserObject()
        )
        # read num_of_threads from environment
        num_of_threads = decouple.config("THREADPOOL.NUMOFTHREADS", num_of_threads, cast=int)
        self.num_of_threads = num_of_threads if num_of_threads is not None else 5
        # create worker threads
        for _ in range(num_of_threads):
            self.add_thread(target=self.callback)
        # register atexit function
        atexit.register(self.stop_threads)

    # -------------------------------------------------------------
    # STOP THREADS METHOD DEFINITION
    # -------------------------------------------------------------
    def add_thread(self, target: types.FunctionType) -> None:
        """
        Adds a thread to the thread pool.
        """
        thread = threading.Thread(
            target=target,
            args=(
                len(self.thread_list),
                self.thread_local_data,
                self.input_queue,
                self.output_queue,
                self.thread_pool_user_objects,
            ),
            daemon=True,
        )
        # start the thread
        thread.start()
        # add the thread to the thread list
        self.thread_list.append(thread)

    # -------------------------------------------------------------
    # STOP THREADS METHOD DEFINITION
    # -------------------------------------------------------------
    def stop_threads(self) -> None:
        """
        Stops all the threads in the thread pool by putting a stop signal
        in the input queue.
        """
        for _ in range(len(self.thread_list)):
            self.input_queue.put("##STOP##")

    # -------------------------------------------------------------
    # REGISTER MESSAGE HANDLER METHOD DEFINITION
    # -------------------------------------------------------------
    def register_message_handler(self, msg_type: str, msg_handler: ThreadPoolUserObject) -> None:
        """
        Registers a message handler for a specific message type.
        """
        self.thread_pool_user_objects[msg_type] = msg_handler

    # -------------------------------------------------------------
    # GET MESSAGE HANDLER METHOD DEFINITION
    # -------------------------------------------------------------
    def get_message_handler(self, msg_type: str) -> ThreadPoolUserObject:
        """
        Get message handler for a specific message type.
        """
        if msg_type in self.thread_pool_user_objects:
            return self.thread_pool_user_objects[msg_type]
        else:
            return self.thread_pool_user_objects["default"]

    # -------------------------------------------------------------
    # CALLBACK METHOD DEFINITION
    # -------------------------------------------------------------
    def callback(
        self,
        _,
        thread_local_data,
        input_queue,
        output_queue,
        thread_pool_user_objects,
    ) -> None:
        """
        Executes the callback function for each item in the input queue.

        Args:
            thread_num (int): The number of the thread.
            thread_local_data (Any): The thread local data.
            input_queue (Queue): The input queue containing the items to
                process.
            output_queue (Queue): The output queue to store the results.
            thread_pool_function (ThreadPoolFunction): The thread pool
                function object.

        Returns:
            None
        """
        for message in iter(input_queue.get, "##STOP##"):
            # if item from queue is no ThreadPoolMessage, raise an exception
            if not isinstance(message, ThreadPoolMessage):
                raise ValueError("Message is not a ThreadPoolMessage")
            # retry until successful or max retries reached
            result = 1
            for _ in range(message.retry):
                with TimerThread(message.timeout, threading.get_ident()):
                    # init variables
                    result = 1
                    try:
                        thread_pool_user_object = (
                            thread_pool_user_objects[message.type]
                            if message.type in thread_pool_user_objects
                            else thread_pool_user_objects["default"]
                        )
                        # call the callable function
                        result = thread_pool_user_object.callable_function(
                            thread_local_data,
                            input_queue,
                            output_queue,
                            message,
                        )
                    except TimeoutError:
                        # Log the timeout message instead of printing
                        print(f"##TIMEOUT## - {message.content}")
                    except Exception as e:
                        print(f"##EXCEPTION## - {message.content} - {e}")
                # check if the task was successful
                if result in (None, 0):
                    break

            # if not successful, put an item into the output queue
            if result not in (None, 0):
                # Log the failure message instead of printing
                print(f"##FAILURE## - {message.content}")
                output_queue.put(("##FAILURE##", message.content))
            else:
                # check to call the hook function
                if message.hook is not None:
                    message.hook(thread_local_data, input_queue, output_queue, message.content)

            # signal that the task is done
            input_queue.task_done()

    # -------------------------------------------------------------
    # GET INPUT QUEUE METHOD DEFINITION
    # -------------------------------------------------------------
    def get_input_queue(self) -> queue.Queue | None:
        """
        Returns the input queue of the thread pool.

        Returns:
            The input queue of the thread pool.
        """
        return self.input_queue

    # -------------------------------------------------------------
    # GET OUTPUT QUEUE METHOD DEFINITION
    # -------------------------------------------------------------
    def get_output_queue(self) -> queue.Queue | None:
        """
        Returns the output queue of the thread pool.

        Returns:
            The output queue of the thread pool.
        """
        return self.output_queue

    # -------------------------------------------------------------
    # GET DATAFRAMES FROM OUTPUT QUEUE AND JOIN THEM INTO DICTIONARY
    # -------------------------------------------------------------
    def get_dataframes_from_output_queue(self) -> dict:
        """
        Receive a list of DataFrames from the EOD2PD message handler.

        Returns
        -------
        list
            A list of DataFrames from the EOD2PD message handler.
        """
        # get the list of dataframes
        result = {}
        while not self.get_output_queue().empty():
            result = result | self.get_output_queue().get()
        return result


class TimerThread:
    """
    A class representing a timer thread that raises a RuntimeError in the
    thread with the specified thread_id to terminate it after a specified
    timeout.
    """

    timeout: int | None = None
    thread_id: int | None = None
    timer: threading.Thread | None = None

    def __init__(self, timeout: int, thread_id: int) -> None:
        """
        Initializes a TimerThread object with the specified timeout and
        thread_id.

        Args:
            timeout (float): The timeout value in seconds.
            thread_id (int): The ID of the thread to terminate.
        """
        self.timeout = timeout
        self.thread_id = thread_id
        self.timer = threading.Timer(
            interval=self.timeout,
            function=self.timeout_thread,
            args=[],
        )

    def __enter__(self):
        """
        Starts the timer thread.
        """
        self.timer.start()

    def __exit__(self, _type, _value, _traceback):
        """
        Cancels the timer thread.
        """
        self.timer.cancel()

    def timeout_thread(self):
        """
        Raises a TimeoutError in the thread with the specified thread_id to
        terminate it.
        """
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(self.thread_id),
            ctypes.py_object(TimeoutError),
        )

==> basefunctions-0.3.7-py3-none-any.whl: basefunctions/threadpool.py <==
"""
# =============================================================================
#
#  Licensed Materials, Property of Ralph Vogl, Munich
#
#  Project : eod2pd
#
#  Copyright (c) by Ralph Vogl
#
#  All rights reserved.
#
#  Description:
#
#  a simple thread pool library to execute commands in a multithreaded
#  environment
#
#  The thread pool lib supports a timeout mechanism when executing the
#  commands. Specify the timeout value in seconds when creating the thread
#  pool. If the execution time of the command exceeds the timeout value,
#  a timout exception will occur and the command will be retried.
#  The number of retries can also be specified when creating the thread pool.
#  If the command wasn't successful after the specified number of retries,
#  the command will be considered as failed. In this case, a message
#  '##FAILURE##' will be put into the output queue and the item will be put
#  back into the input queue. If the command was successful, the result of
#  the command will be put into the output queue.
# =============================================================================
"""

# pylint: disable=W0718

# -------------------------------------------------------------
# IMPORTS
# -------------------------------------------------------------
import atexit
import ctypes
import queue
import threading
import types
from dataclasses import dataclass
from typing import Any, Callable, List

import decouple
import numpy as np
import pandas as pd

# -------------------------------------------------------------
# DEFINITIONS REGISTRY
# -------------------------------------------------------------

# -------------------------------------------------------------
# DEFINITIONS
# -------------------------------------------------------------


# -------------------------------------------------------------
# VARIABLE DEFINTIONS
# -------------------------------------------------------------


# -------------------------------------------------------------
#  CLASS DEFINITIONS
# -------------------------------------------------------------


@dataclass
class ThreadPoolMessage:
    """
    This class represents a message that can be sent to the thread pool.
    """

    type: str
    retry: int = 3
    timeout: int = 5
    content: Any = None
    hook: Callable[[Any, queue.Queue, queue.Queue, Any], None] = None


def create_threadpool_message(
    _type: str, content: Any, retry: int = 3, timeout: int = 5, hook: Callable = None
) -> ThreadPoolMessage:
    """
    Creates a ThreadPoolMessage object with the specified type and content.

    Args:
        _type (str): The type of the message.
        content (Any): The content of the message.
        retry (int): The number of retries for the message.
        timeout (int): The timeout value for the message.
        hook(Callable): The hook function to call after the message was processed.

    Returns:
        ThreadPoolMessage: The created ThreadPoolMessage object.
    """
    return ThreadPoolMessage(_type, retry, timeout, content, hook)


# pylint too-few-public-methods
class ThreadPoolUserObject:
    """
    This class represents a user object which contains the callable_function
    that can be executed by a thread pool.

    Methods
    -------
    callableFunction(outputQueue, item)
        This function will be called from the thread pool in order to execute
        a specific command. The command is provided in the item parameter
        and is application specific. After processing the command, the result
        of the command can be put into the outputQueue.

    Attributes
    ----------
    None
    """

    def callable_function(
        self,
        thread_local_data: Any,
        input_queue: queue.Queue,
        output_queue: queue.Queue,
        message: ThreadPoolMessage,
    ) -> int:
        """
        This function will be called from the thread pool in order to execute
        a specific command. The command is provided in the message parameter
        and is application specific. After processing the command, the result
        of the command can be put into the output_queue. If there are any
        other commands that needs to be executed, they can be put into the
        input_queue.

        Parameters
        ----------
        thread_local_data : Any
            The thread local data.
        input_queue : queue.LifoQueue
            The input queue for receiving tasks.
        output_queue : queue.Queue
            The queue to put the result of the task.
        message : object
            The message containing the command to execute.

        Returns
        -------
        int
            The return code of the task.
            Return 0 if successful, otherwise any other value.
        """
        raise NotImplementedError


class ThreadPool:
    """
    A thread pool implementation for executing tasks concurrently.

    This class manages a pool of threads that can execute tasks in parallel.
    It provides a convenient way to distribute work across multiple threads
    and process the results asynchronously.

    Attributes:
        thread_pool (list): A list of threads in the thread pool.
        input_queue (Queue): The input queue for receiving tasks.
        output_queue (Queue): The output queue for storing results.
        thread_pool_user_objects (dict): A dictionary of thread pool user

    Methods:
        add_thread(target)
            Adds a thread to the thread pool.
        stop_threads()
            Stops all the threads in the thread pool by putting a stop signal
            in the input queue.
        register_message_handler(msg_type, msg_handler)
            Registers a message handler for a specific message type.
        get_message_handler(msg_type)
            Get message handler for a specific message type.
        callback(thread_num, thread_local_data, input_queue, output_queue,
                thread_pool_user_objects)
            Executes the callback function for each item in the input queue.
        get_input_queue()
            Returns the input queue of the thread pool.
        get_output_queue()
            Returns the output queue of the thread pool.
        get_dataframes_from_output_queue()
            Receive a list of DataFrames from the output queue.
    """

    # -------------------------------------------------------------
    # VARIABLE DEFINTIONS
    # -------------------------------------------------------------
    thread_list: List[threading.Thread] = []
    input_queue: queue.Queue = queue.Queue()
    output_queue: queue.Queue = queue.Queue()
    thread_pool_user_objects: dict = {}
    thread_local_data = threading.local()

    # -------------------------------------------------------------
    # INIT METHOD DEFINITION
    # -------------------------------------------------------------
    def __init__(
        self,
        default_thread_pool_user_object: ThreadPoolUserObject,
        num_of_threads: int = 5,
    ) -> None:
        # save the default thread pool user object
        self.thread_pool_user_objects["default"] = (
            default_thread_pool_user_object
            if default_thread_pool_user_object is not None
            else ThreadPoolUserObject()
        )
        # read num_of_threads from environment
        num_of_threads = decouple.config("THREADPOOL.NUMOFTHREADS", num_of_threads, cast=int)
        self.num_of_threads = num_of_threads if num_of_threads is not None else 5
        # create worker threads
        for _ in range(num_of_threads):
            self.add_thread(target=self.callback)
        # register atexit function
        atexit.register(self.stop_threads)

    # -------------------------------------------------------------
    # STOP THREADS METHOD DEFINITION
    # -------------------------------------------------------------
    def add_thread(self, target: types.FunctionType) -> None:
        """
        Adds a thread to the thread pool.
        """
        thread = threading.Thread(
            target=target,
            args=(
                len(self.thread_list),
                self.thread_local_data,
                self.input_queue,
                self.output_queue,
                self.thread_pool_user_objects,
            ),
            daemon=True,
        )
        # start the thread
        thread.start()
        # add the thread to the thread list
        self.thread_list.append(thread)

    # -------------------------------------------------------------
    # STOP THREADS METHOD DEFINITION
    # -------------------------------------------------------------
    def stop_threads(self) -> None:
        """
        Stops all the threads in the thread pool by putting a stop signal
        in the input queue.
        """
        for _ in range(len(self.thread_list)):
            self.input_queue.put("##STOP##")

    # -------------------------------------------------------------
    # REGISTER MESSAGE HANDLER METHOD DEFINITION
    # -------------------------------------------------------------
    def register_message_handler(self, msg_type: str, msg_handler: ThreadPoolUserObject) -> None:
        """
        Registers a message handler for a specific message type.
        """
        self.thread_pool_user_objects[msg_type] = msg_handler

    # -------------------------------------------------------------
    # GET MESSAGE HANDLER METHOD DEFINITION
    # -------------------------------------------------------------
    def get_message_handler(self, msg_type: str) -> ThreadPoolUserObject:
        """
        Get message handler for a specific message type.
        """
        if msg_type in self.thread_pool_user_objects:
            return self.thread_pool_user_objects[msg_type]
        else:
            return self.thread_pool_user_objects["default"]

    # -------------------------------------------------------------
    # CALLBACK METHOD DEFINITION
    # -------------------------------------------------------------
    def callback(
        self,
        _,
        thread_local_data,
        input_queue,
        output_queue,
        thread_pool_user_objects,
    ) -> None:
        """
        Executes the callback function for each item in the input queue.

        Args:
            thread_num (int): The number of the thread.
            thread_local_data (Any): The thread local data.
            input_queue (Queue): The input queue containing the items to
                process.
            output_queue (Queue): The output queue to store the results.
            thread_pool_function (ThreadPoolFunction): The thread pool
                function object.

        Returns:
            None
        """
        for message in iter(input_queue.get, "##STOP##"):
            # if item from queue is no ThreadPoolMessage, raise an exception
            if not isinstance(message, ThreadPoolMessage):
                raise ValueError("Message is not a ThreadPoolMessage")
            # retry until successful or max retries reached
            result = 1
            for _ in range(message.retry):
                with TimerThread(message.timeout, threading.get_ident()):
                    # init variables
                    result = 1
                    try:
                        thread_pool_user_object = (
                            thread_pool_user_objects[message.type]
                            if message.type in thread_pool_user_objects
                            else thread_pool_user_objects["default"]
                        )
                        # call the callable function
                        result = thread_pool_user_object.callable_function(
                            thread_local_data,
                            input_queue,
                            output_queue,
                            message,
                        )
                    except TimeoutError:
                        # Log the timeout message instead of printing
                        print(f"##TIMEOUT## - {message.content}")
                    except Exception as e:
                        print(f"##EXCEPTION## - {message.content} - {e}")
                # check if the task was successful
                if result in (None, 0):
                    break

            # if not successful, put an item into the output queue
            if result not in (None, 0):
                # Log the failure message instead of printing
                print(f"##FAILURE## - {message.content}")
                output_queue.put(("##FAILURE##", message.content))
            else:
                # check to call the hook function
                if message.hook is not None:
                    message.hook(thread_local_data, input_queue, output_queue, message.content)

            # signal that the task is done
            input_queue.task_done()

    # -------------------------------------------------------------
    # GET INPUT QUEUE METHOD DEFINITION
    # -------------------------------------------------------------
    def get_input_queue(self) -> queue.Queue | None:
        """
        Returns the input queue of the thread pool.

        Returns:
            The input queue of the thread pool.
        """
        return self.input_queue

    # -------------------------------------------------------------
    # GET OUTPUT QUEUE METHOD DEFINITION
    # -------------------------------------------------------------
    def get_output_queue(self) -> queue.Queue | None:
        """
        Returns the output queue of the thread pool.

        Returns:
            The output queue of the thread pool.
        """
        return self.output_queue

    # -------------------------------------------------------------
    # GET DATAFRAMES FROM OUTPUT QUEUE AND JOIN THEM INTO DICTIONARY
    # -------------------------------------------------------------
    def get_dataframes_from_output_queue(
        self, dict_result: bool = True, inner_dates: bool = False
    ) -> pd.DataFrame | dict:
        """
        Receive a list of DataFrames from the EOD2PD message handler.

        Parameters
        ----------
        dict_result : bool
            If True, the result will be a dictionary with the message type as key.
            If False, the result will be a list of DataFrames.
        inner_dates : bool
            If False, the DataFrames will be filled with None for missing dates.
            If True, the DataFrames will be aligned to the newest date.

        Returns
        -------
        dict or DataFrame
            A dictionary of DataFrames if dict_result is True or if there are multiple DataFrames.
            Otherwise, returns a single DataFrame.
        """
        # TODO: inner_dates in jedem Fall berücksichtigen

        result = {}
        while not self.get_output_queue().empty():
            result.update(self.get_output_queue().get())
        if dict_result:
            return result
        else:
            if len(result) > 1:
                return self.create_multiindex_dataframe(result=result, inner_dates=inner_dates)
            else:
                return result[list(result.keys())[0]]

    def create_multiindex_dataframe(self, result: dict, inner_dates=False) -> pd.DataFrame:
        """
        Create a multi-index DataFrame from a dictionary of DataFrames.

        Parameters:
        -----------
        result : dict
            A dictionary of DataFrames.
        inner_dates : bool
            If False, the DataFrames will be filled with None for missing dates.
            If True, the DataFrames will be aligned to the newest date.

        Returns:
        --------
        pd.DataFrame
            A multi-index DataFrame containing the DataFrames from the dictionary.
        """
        # align all DataFrames to start and end with the date range
        keys = sorted(result.keys())
        values = [result[key] for key in keys]

        # Concatenate aligned DataFrames with MultiIndex columns
        result = pd.concat(
            values,
            axis=1,
            join="inner" if inner_dates else "outer",
            sort=True,
            keys=[key.upper() for key in keys],
        )

        return result


class TimerThread:
    """
    A class representing a timer thread that raises a RuntimeError in the
    thread with the specified thread_id to terminate it after a specified
    timeout.
    """

    timeout: int | None = None
    thread_id: int | None = None
    timer: threading.Thread | None = None

    def __init__(self, timeout: int, thread_id: int) -> None:
        """
        Initializes a TimerThread object with the specified timeout and
        thread_id.

        Args:
            timeout (float): The timeout value in seconds.
            thread_id (int): The ID of the thread to terminate.
        """
        self.timeout = timeout
        self.thread_id = thread_id
        self.timer = threading.Timer(
            interval=self.timeout,
            function=self.timeout_thread,
            args=[],
        )

    def __enter__(self):
        """
        Starts the timer thread.
        """
        self.timer.start()

    def __exit__(self, _type, _value, _traceback):
        """
        Cancels the timer thread.
        """
        self.timer.cancel()

    def timeout_thread(self):
        """
        Raises a TimeoutError in the thread with the specified thread_id to
        terminate it.
        """
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(self.thread_id),
            ctypes.py_object(TimeoutError),
        )
